/* This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://www.wtfpl.net/ for more details. */

#ifndef GLOBALS_H_INCLUDED
#define GLOBALS_H_INCLUDED

#include "config.h"
#include <iostream>
#include <vector>

namespace Globals
{
    extern GalaxyConfig galaxyConf;
    extern Config globalConfig;
    extern std::vector<std::string> vOwnedGamesIds;
}

#endif // GLOBALS_H_INCLUDED

